import { BaseError } from 'make-error';
import Container from '../Container';
import { Mod } from '../../mods';
import { Tag, BaseItemTypeProps } from '../../schema';
import MetaData from '../../util/MetaData';
import Stat from '../../calculator/Stat';
import { AtlasModifier } from './atlasModifier';
import Component from './Component';
import ItemAffixes from './components/Affixes';
import { Sockets, Builder as SocketsBuilder } from './components/Sockets';
import { Name, Builder as NameBuilder } from './components/Name';
import { Builder as RarityBuilder, Rarity } from './components/Rarity';
import Implicits from './components/Implicits';
import { Requirements, Builder as RequirementsBuilder } from './components/Requirements';
import { Properties, Builder as PropertiesBuilder } from './components/properties';
import { AugmentableValue } from './util';
export interface ItemProps {
    readonly atlas_modifier: AtlasModifier;
    readonly corrupted: boolean;
    readonly item_level: number;
    readonly mirrored: boolean;
}
export declare class UnacceptedMod extends BaseError {
    constructor();
}
export interface Builder {
    /**
     * explicits of the item
     */
    affixes: Mod[];
    baseitem: BaseItemTypeProps;
    implicits: Mod[];
    /**
     * reflection of the baseitem
     */
    meta_data: MetaData;
    name: NameBuilder;
    props: ItemProps;
    properties: PropertiesBuilder;
    rarity: RarityBuilder;
    requirements: RequirementsBuilder;
    sockets: SocketsBuilder;
}
/**
 * an Item in Path of Exile
 */
export default class Item implements Container<Mod> {
    static buildMetaData(baseitem: BaseItemTypeProps): MetaData;
    /**
     * creates a new item from the baseitem
     * @param baseitem
     */
    static build(baseitem: BaseItemTypeProps): Item;
    static fromBuilder(builder: Builder): Item;
    affixes: ItemAffixes;
    baseitem: BaseItemTypeProps;
    implicits: Implicits;
    meta_data: MetaData;
    name: Name & Component<Item, NameBuilder>;
    props: ItemProps;
    properties: Properties & Component<Item, any>;
    rarity: Rarity<Item> & Component<Item, RarityBuilder>;
    requirements: Requirements & Component<Item, RequirementsBuilder>;
    sockets: Sockets<Item> & Component<Item, SocketsBuilder>;
    /**
     * Use Item#build
     *
     * @private
     * @param builder
     */
    constructor(builder: Builder);
    withMutations(mutate: (builder: Builder) => Builder): this;
    builder(): Builder;
    /**
     * returns tags of item + tags from mods
     */
    getTags(): Tag[];
    readonly mods: Mod[];
    asArray(): Mod[];
    /**
     * decides where to add the mod (explicit, implicit)
     * throws if it could not decide where to put it
     * @param other
     */
    addMod(other: Mod): this;
    /**
     * removed this mod either from implicit or explicit
     *
     * if that mod fiths into neither category it throws
     * @param other
     */
    removeMod(other: Mod): this;
    /**
     * removes explicits
     */
    removeAllMods(): this;
    hasMod(other: Mod): boolean;
    hasModGroup(other: Mod): boolean;
    hasRoomFor(other: Mod): boolean;
    indexOfModWithId(id: string): number;
    maxModsOfType(other: Mod): number;
    inDomainOf(mod_domain: number): boolean;
    level(): number;
    any(): boolean;
    /**
     * merge of implicit and explicit stats
     */
    stats(): {
        [key: string]: Stat;
    };
    removeAllImplicits(): this;
    setProperty(prop: keyof ItemProps, value: any): this;
    /**
     * sets the corrupted property on the item or throws if it is already corrupted
     */
    corrupt(): this;
    /**
     * sets the mirror property on the item or throws if it is already mirrored
     */
    mirror(): this;
    isElderItem(): boolean;
    /**
     * returns an item that can have elder mods
     *
     * this does not remove existing shaper mods
     */
    asElderItem(): this;
    isSHaperItem(): boolean;
    /**
     * returns an item that can have shaper mods
     *
     * this does not remove existing elder mods
     */
    asShaperItem(): this;
    removeAtlasModifier(): this;
    /**
     * augments a given {value} with the local stats
     * @param value
     * @param classification
     */
    computeValue(value: number, classification: string[]): AugmentableValue;
    private mutateAffixes;
    private addAffix;
    private removeAffix;
    private mutateImplicits;
    private addImplicit;
    private removeImplicit;
    private asAtlasModifier;
}
