"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var errors_1 = require("./errors");
var types_1 = require("./types");
/**
 * item component for sockets
 */
var ItemSockets = /** @class */ (function () {
    function ItemSockets(item, builder) {
        this.parent = item;
        this.groups = builder.groups;
        this.sockets = builder.sockets;
    }
    ItemSockets.prototype.builder = function () {
        return {
            groups: this.groups,
            sockets: this.sockets,
        };
    };
    /**
     * @deprecated
     */
    ItemSockets.prototype.max = function () {
        console.warn('poe-mods DEPRECATION: use Sockets.maxSockets() instead of Sockets.max()');
        return this.maxSockets();
    };
    /**
     * calculates the maximum amount of sockets possible on this item
     *
     * TODO: what about Corroded Blades or other similar 1x4 Items. Confirm
     * that they also only can have max 4 sockets like Rods or act like small_Staff
     */
    ItemSockets.prototype.maxSockets = function () {
        var by_stats = this.maxOverride();
        // tags take priority
        if (by_stats != null) {
            return by_stats;
        }
        else {
            return Math.min(this.maxByDimensions(), this.maxByLevel(), this.maxByMetaData());
        }
    };
    ItemSockets.prototype.count = function () {
        return this.sockets.length;
    };
    ItemSockets.prototype.maxLinks = function () {
        // Math.max of empty array returns -inf
        if (this.groups.length === 0) {
            return 0;
        }
        else {
            return Math.max.apply(Math, this.groups.map(function (g) { return g.length; }));
        }
    };
    ItemSockets.prototype.isLinked = function (a, b) {
        var group_a = this.groupOf(a);
        // -1 hints that this socket is part of no group and therefore cannot be linked
        if (group_a === -1) {
            return false;
        }
        var group_b = this.groupOf(b);
        return group_a === group_b;
    };
    /**
     * links the specified socket indices
     * @param pairs consider args as pairs so be sure to have even argument length
     */
    ItemSockets.prototype.link = function () {
        var pairs = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            pairs[_i] = arguments[_i];
        }
        if (pairs.length % 2) {
            throw new errors_1.ArgumentError('#links() require pairs of sockets. An odd number of sockets was given.');
        }
        return this.parent.withMutations(function (builder) {
            var _a = builder.sockets, groups = _a.groups, sockets = _a.sockets;
            // for each pair link the pair but throw if the specified
            // sockets dont exist
            var new_groups = chunk(pairs, 2).reduce(function (acc, _a) {
                var a = _a[0], b = _a[1];
                if (sockets[a] === undefined || sockets[b] === undefined) {
                    throw new errors_1.LinkNonExistingSockets();
                }
                return link(acc, a, b);
            }, groups);
            if (new_groups === groups) {
                return builder;
            }
            else {
                return __assign({}, builder, { sockets: __assign({}, builder.sockets, { groups: new_groups }) });
            }
        });
    };
    /**
     * set the sockets to {n}
     *
     * set force to true to ignore socket limit
     * set keep_links to false if the new sockets should all be unlinked
     * newSocket is a callback which is called for every new socket. By default
     * a white socket is created.
     *
     * @param n
     * @param options
     */
    ItemSockets.prototype.socket = function (n, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        var _a = options.newSocket, newSocket = _a === void 0 ? defaultNewSocket : _a, _b = options.keep_links, keep_links = _b === void 0 ? true : _b, _c = options.force, force = _c === void 0 ? false : _c;
        var max_count = this.maxSockets();
        if (!force && n > max_count) {
            throw new errors_1.SocketOverflow(max_count);
        }
        return this.parent.withMutations(function (builder) {
            var count = _this.count();
            if (count === n) {
                return builder;
            }
            else {
                var new_sockets = Array.from({ length: n }).map(function (_, index) {
                    return (builder.sockets.sockets[index] || newSocket(_this.parent, index));
                });
                // slice socket indices away that are beyond the new count
                var new_groups = keep_links
                    ? builder.sockets.groups
                        .map(function (group) { return group.filter(function (_, index) { return index < n; }); })
                        .filter(function (group) { return group.length > 0; })
                    : [];
                return __assign({}, builder, { sockets: {
                        sockets: new_sockets,
                        groups: new_groups,
                    } });
            }
        });
    };
    ItemSockets.prototype.colors = function () {
        return this.sockets.map(function (_a) {
            var color = _a.color;
            return color;
        });
    };
    ItemSockets.prototype.color = function (color) {
        var indices = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            indices[_i - 1] = arguments[_i];
        }
        return this.colorBatch({ color: color, sockets: indices });
    };
    ItemSockets.prototype.colorBatch = function () {
        var batches = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            batches[_i] = arguments[_i];
        }
        if (this.batchesHaveNonExistingSockets(batches)) {
            throw new errors_1.ColorNonExistingSocket();
        }
        // SocketId => NewSocketColor
        var colors = batches.reduce(function (acc, batch) {
            return batch.sockets.reduce(function (acc_batch, socket_index) {
                return acc_batch.set(socket_index, batch.color);
            }, acc);
        }, new Map());
        return this.mapColors(function (socket, index) {
            var new_color = colors.get(index);
            // dont recolor this
            if (new_color === undefined) {
                return socket;
            }
            else {
                return __assign({}, socket, { color: new_color });
            }
        });
    };
    /**
     *
     * @param fn if fn returns a referentially equal Socket no new item is returned
     */
    ItemSockets.prototype.mapColors = function (fn) {
        var _this = this;
        return this.parent.withMutations(function (builder) {
            var changed = false;
            var new_sockets = builder.sockets.sockets.map(function (socket, index) {
                var new_socket = fn(socket, index, _this.parent);
                if (new_socket.color !== socket.color) {
                    changed = true;
                }
                return new_socket;
            });
            if (!changed) {
                return builder;
            }
            else {
                return __assign({}, builder, { sockets: __assign({}, builder.sockets, { 
                        // copy array
                        sockets: new_sockets }) });
            }
        });
    };
    ItemSockets.prototype.any = function () {
        return this.count() > 0;
    };
    /**
     * a string representation of the sockets, links and colors
     *
     * each char represents a color shorthand, dashes means the adjacent sockets
     * are linked, whitespace instead means no link
     */
    ItemSockets.prototype.toString = function () {
        var _this = this;
        return this.sockets
            .map(function (socket, index) {
            // last socket?
            if (index + 1 === _this.sockets.length) {
                return "" + socket.color;
            }
            else {
                // linked with the next socket?
                var is_linked = _this.isLinked(index, index + 1);
                return "" + socket.color + (is_linked ? '-' : ' ');
            }
        })
            .join('');
    };
    /**
     * serialized format which includes group information
     */
    ItemSockets.prototype.toJson = function () {
        var _this = this;
        return this.sockets.map(function (socket, index) {
            var group = _this.groupOf(index);
            return __assign({}, socket, { group: group });
        });
    };
    ItemSockets.prototype.groupOf = function (socket_index) {
        return groupOf(this.groups, socket_index);
    };
    ItemSockets.prototype.maxByMetaData = function () {
        var meta_data = this.parent.meta_data;
        if (meta_data.isA('AbstractShield')) {
            return 3;
        }
        else if (meta_data.isA('AbstractArmour')) {
            return 6;
        }
        else if (meta_data.isA('AbstractOneHandWeapon')) {
            return 3;
        }
        else if (meta_data.isA('AbstractFishingRod')) {
            return 4;
        }
        else if (meta_data.isA('AbstractTwoHandWeapon')) {
            return 6;
        }
        else if (meta_data.isA('Equipment')) {
            return 0;
        }
        else {
            throw new Error("Can't determine sockes from meta data for " + meta_data.clazz);
        }
    };
    ItemSockets.prototype.maxByLevel = function () {
        var props = this.parent.props;
        if (props.item_level <= 1) {
            return 2;
        }
        else if (props.item_level <= 2) {
            return 3;
        }
        else if (props.item_level <= 25) {
            return 4;
        }
        else if (props.item_level <= 35) {
            return 5;
        }
        else {
            return 6;
        }
    };
    ItemSockets.prototype.maxByDimensions = function () {
        var _a = this.parent.baseitem, width = _a.width, height = _a.height;
        return width * height;
    };
    ItemSockets.prototype.maxOverride = function () {
        var stats = this.parent.stats();
        var tags = this.parent.getTags();
        if (stats.local_has_X_sockets != null) {
            return stats.local_has_X_sockets.values.max;
        }
        else if (tags.find(function (id) { return id === 'small_staff'; }) !== undefined) {
            return 3;
        }
        return undefined;
    };
    /**
     * true if some batch tries to color an index that is out of bounds
     * @param batches
     */
    ItemSockets.prototype.batchesHaveNonExistingSockets = function (batches) {
        var _this = this;
        return batches.some(function (batch) {
            return batch.sockets.some(function (index) { return _this.sockets[index] === undefined; });
        });
    };
    return ItemSockets;
}());
exports.default = ItemSockets;
function defaultNewSocket(item) {
    return {
        color: types_1.SocketColor.white,
    };
}
/**
 * takes a flat array and transforms it into an array of arrays of {size}
 * @param arr
 * @param size
 */
function chunk(arr, size) {
    return Array.from({ length: Math.ceil(arr.length / size) }, function (_, i) {
        return arr.slice(i * size, (i + 1) * size);
    });
}
function unique(arr) {
    return Array.from(new Set(arr));
}
// staic helpers
/**
 * ind the index of a group that holds {socket_index}
 * @param groups
 * @param socket_index
 */
function groupOf(groups, socket_index) {
    return groups.findIndex(function (group) { return group.includes(socket_index); });
}
function link(groups, a, b) {
    var _a;
    var group_a = groupOf(groups, a);
    var group_b = groupOf(groups, b);
    // already linked?
    if (group_a !== -1 && group_a === group_b) {
        return groups;
    }
    // keep every group that is not within [a, b] as-is
    var unaffected_groups = groups.filter(function (group) {
        var overlapping = group.find(function (socket) { return a <= socket && socket <= b; });
        return overlapping === undefined;
    });
    // and add a new group [a..b] merged with the original groups
    var merged_group = (_a = Array.from({ length: b - a + 1 }, function (_, i) { return i + a; })).concat.apply(_a, (groups[group_a] || []).concat((groups[group_b] || [])));
    return unaffected_groups.concat([unique(merged_group)]);
}
