var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
import { BaseError } from 'make-error';
import { Mod } from '../../mods';
import MetaData from '../../util/MetaData';
import Stat from '../../calculator/Stat';
import Value from '../../calculator/Value';
import atlasModifier, { AtlasModifier, tagsWithModifier as tagsWithAtlasModifier, } from './atlasModifier';
import ItemAffixes from './components/Affixes';
import ItemSockets from './components/Sockets';
import ItemName from './components/Name';
import ItemRarity, { RarityKind, } from './components/Rarity';
import Implicits from './components/Implicits';
import ItemRequirements from './components/Requirements';
import { build as buildProperties, } from './components/properties';
var UnacceptedMod = /** @class */ (function (_super) {
    __extends(UnacceptedMod, _super);
    function UnacceptedMod() {
        return _super.call(this, 'Unacceptable mods passed to this Container') || this;
    }
    return UnacceptedMod;
}(BaseError));
export { UnacceptedMod };
var shallowEqual = function (a, b) {
    return a === b || Object.keys(a).every(function (key) { return a[key] === b[key]; });
};
/**
 * an Item in Path of Exile
 */
var Item = /** @class */ (function () {
    /**
     * Use Item#build
     *
     * @private
     * @param builder
     */
    function Item(builder) {
        this.baseitem = builder.baseitem;
        this.props = builder.props;
        this.meta_data = builder.meta_data;
        // components
        this.name = new ItemName(this, builder.name);
        this.properties = buildProperties(this, builder.properties);
        this.rarity = new ItemRarity(this, builder.rarity);
        this.requirements = new ItemRequirements(this, builder.requirements);
        this.sockets = new ItemSockets(this, builder.sockets);
        this.affixes = new ItemAffixes(this, builder.affixes);
        this.implicits = new Implicits(this, builder.implicits);
    }
    Item.buildMetaData = function (baseitem) {
        var clazz = String(baseitem.inherits_from.split(/[\\/]/).pop());
        return MetaData.build(clazz);
    };
    /**
     * creates a new item from the baseitem
     * @param baseitem
     */
    Item.build = function (baseitem) {
        var meta_data = Item.buildMetaData(baseitem);
        var implicits = baseitem.implicit_mods.map(function (mod_props) { return new Mod(mod_props); });
        return new Item({
            affixes: [],
            baseitem: baseitem,
            implicits: implicits,
            meta_data: meta_data,
            name: 'Random Name',
            /**
             * miscellaneous props
             */
            props: {
                atlas_modifier: atlasModifier(baseitem),
                corrupted: false,
                item_level: 100,
                mirrored: false,
            },
            /**
             * calculation related props
             */
            properties: { quality: 0 },
            rarity: RarityKind.normal,
            requirements: baseitem.component_attribute_requirement,
            /**
             * the sockets of the item
             */
            sockets: {
                sockets: [],
                groups: [],
            },
        });
    };
    Item.fromBuilder = function (builder) {
        return new Item(builder);
    };
    Item.prototype.withMutations = function (mutate) {
        var prev = this.builder();
        var mutated = mutate(prev);
        if (shallowEqual(prev, mutated)) {
            return this;
        }
        else {
            // @ts-ignore
            return this.constructor.fromBuilder(mutated);
        }
    };
    Item.prototype.builder = function () {
        return {
            affixes: this.affixes.mods,
            baseitem: this.baseitem,
            implicits: this.implicits.mods,
            meta_data: this.meta_data,
            name: this.name.builder(),
            props: this.props,
            properties: this.properties.builder(),
            rarity: this.rarity.builder(),
            requirements: this.requirements.builder(),
            sockets: this.sockets.builder(),
        };
    };
    // Taggable Implementation
    /**
     * returns tags of item + tags from mods
     */
    Item.prototype.getTags = function () {
        return this.meta_data.props.tags.concat(this.baseitem.tags, this.implicits.getTags(), this.affixes.getTags());
    };
    Object.defineProperty(Item.prototype, "mods", {
        // Container implementtion
        get: function () {
            return this.implicits.mods.concat(this.affixes.mods);
        },
        enumerable: true,
        configurable: true
    });
    Item.prototype.asArray = function () {
        return this.mods;
    };
    /**
     * decides where to add the mod (explicit, implicit)
     * throws if it could not decide where to put it
     * @param other
     */
    Item.prototype.addMod = function (other) {
        if (other.isAffix()) {
            return this.addAffix(other);
        }
        else if (other.implicitCandidate()) {
            return this.addImplicit(other);
        }
        else {
            throw new UnacceptedMod();
        }
    };
    /**
     * removed this mod either from implicit or explicit
     *
     * if that mod fiths into neither category it throws
     * @param other
     */
    Item.prototype.removeMod = function (other) {
        if (other.isAffix()) {
            return this.removeAffix(other);
        }
        else if (other.implicitCandidate()) {
            return this.removeImplicit(other);
        }
        else {
            throw new UnacceptedMod();
        }
    };
    /**
     * removes explicits
     */
    Item.prototype.removeAllMods = function () {
        return this.mutateAffixes(function (affixes) { return affixes.removeAllMods(); });
    };
    Item.prototype.hasMod = function (other) {
        return this.affixes.hasMod(other) || this.implicits.hasMod(other);
    };
    Item.prototype.hasModGroup = function (other) {
        return (
        // isAffix => this.affixes.hasModGroup(other)
        (!other.isAffix() || this.affixes.hasModGroup(other)) &&
            // implicitCandidate => this.implicits.hasModGroup(other)
            (!other.implicitCandidate() || this.implicits.hasModGroup(other)));
    };
    Item.prototype.hasRoomFor = function (other) {
        return this.affixes.hasRoomFor(other) || this.implicits.hasRoomFor(other);
    };
    Item.prototype.indexOfModWithId = function (id) {
        var affix_index = this.affixes.indexOfModWithId(id);
        var implicit_index = this.implicits.indexOfModWithId(id);
        if (affix_index !== -1) {
            return affix_index;
        }
        else if (implicit_index !== -1) {
            return implicit_index;
        }
        else {
            return -1;
        }
    };
    Item.prototype.maxModsOfType = function (other) {
        if (other.isAffix()) {
            return this.affixes.maxModsOfType(other);
        }
        else if (other.implicitCandidate()) {
            return this.implicits.maxModsOfType(other);
        }
        else {
            throw new UnacceptedMod();
        }
    };
    Item.prototype.inDomainOf = function (mod_domain) {
        return this.affixes.inDomainOf(mod_domain);
    };
    Item.prototype.level = function () {
        return this.props.item_level;
    };
    Item.prototype.any = function () {
        return this.mods.length > 0;
    };
    /**
     * merge of implicit and explicit stats
     */
    Item.prototype.stats = function () {
        // merge implicit stats and affix stats by adding its stats
        var a = this.implicits.stats();
        var b = this.affixes.stats();
        // assume that keys are unique
        return Object.keys(a).concat(Object.keys(b)).reduce(function (stats, key) {
            var _a;
            var left = a[key];
            var right = b[key];
            if (stats[key] != null) {
                // already merged
                return stats;
            }
            else {
                var merged = void 0;
                if (left != null && right != null) {
                    merged = left.add(right.values);
                }
                else if (right != null) {
                    merged = right;
                }
                else if (left != null) {
                    merged = left;
                }
                else {
                    // unreachable
                    throw new Error('unreachable code');
                }
                return __assign({}, stats, (_a = {}, _a[key] = merged, _a));
            }
        }, {});
    };
    // End Container implementation
    Item.prototype.removeAllImplicits = function () {
        return this.mutateImplicits(function (implicits) { return implicits.removeAllMods(); });
    };
    // Begin state managment
    Item.prototype.setProperty = function (prop, value) {
        return this.withMutations(function (builder) {
            var _a;
            return __assign({}, builder, { props: __assign({}, builder.props, (_a = {}, _a[prop] = value, _a)) });
        });
    };
    /**
     * sets the corrupted property on the item or throws if it is already corrupted
     */
    Item.prototype.corrupt = function () {
        if (this.props.corrupted) {
            throw new Error('invalid state: is already corrupted');
        }
        else {
            return this.setProperty('corrupted', true);
        }
    };
    /**
     * sets the mirror property on the item or throws if it is already mirrored
     */
    Item.prototype.mirror = function () {
        if (this.props.mirrored) {
            throw new Error('invalid state: is already mirrored');
        }
        else {
            return this.setProperty('mirrored', true);
        }
    };
    Item.prototype.isElderItem = function () {
        return this.props.atlas_modifier === AtlasModifier.ELDER;
    };
    /**
     * returns an item that can have elder mods
     *
     * this does not remove existing shaper mods
     */
    Item.prototype.asElderItem = function () {
        return this.asAtlasModifier(AtlasModifier.ELDER);
    };
    Item.prototype.isSHaperItem = function () {
        return this.props.atlas_modifier === AtlasModifier.SHAPER;
    };
    /**
     * returns an item that can have shaper mods
     *
     * this does not remove existing elder mods
     */
    Item.prototype.asShaperItem = function () {
        return this.asAtlasModifier(AtlasModifier.SHAPER);
    };
    // returns an item that cant have elder or shaper mods
    // this does not remove existing elder or shaper mods
    Item.prototype.removeAtlasModifier = function () {
        return this.asAtlasModifier(AtlasModifier.NONE);
    };
    // End state
    /**
     * augments a given {value} with the local stats
     * @param value
     * @param classification
     */
    Item.prototype.computeValue = function (value, classification) {
        var base = new Value([value, value], classification);
        return base
            .augmentWith(Object.values(this.stats()))
            .augmentWith([
            new Stat({ id: 'local_quality' }, [
                this.properties.quality,
                this.properties.quality,
            ]),
        ])
            .compute();
    };
    // private
    Item.prototype.mutateAffixes = function (mutate) {
        var _this = this;
        return this.withMutations(function (builder) {
            return __assign({}, builder, { affixes: mutate(_this.affixes).mods });
        });
    };
    Item.prototype.addAffix = function (other) {
        return this.mutateAffixes(function (affixes) { return affixes.addMod(other); });
    };
    Item.prototype.removeAffix = function (other) {
        return this.mutateAffixes(function (affixes) { return affixes.removeMod(other); });
    };
    Item.prototype.mutateImplicits = function (mutate) {
        var _this = this;
        return this.withMutations(function (builder) {
            return __assign({}, builder, { implicits: mutate(_this.implicits).mods });
        });
    };
    Item.prototype.addImplicit = function (other) {
        return this.mutateImplicits(function (implicits) { return implicits.addMod(other); });
    };
    Item.prototype.removeImplicit = function (other) {
        return this.mutateImplicits(function (implicits) { return implicits.removeMod(other); });
    };
    Item.prototype.asAtlasModifier = function (modifier) {
        if (this.props.atlas_modifier === modifier) {
            return this;
        }
        else {
            return this.withMutations(function (builder) {
                return __assign({}, builder, { baseitem: __assign({}, builder.baseitem, { tags: tagsWithAtlasModifier(builder.baseitem, builder.meta_data, modifier) }), props: __assign({}, builder.props, { atlas_modifier: modifier }) });
            });
        }
    };
    return Item;
}());
export default Item;
